-module(recursion).
-export([merge_sort/1,pmerge_sort/1,map/2,pmap/2,reduce/3,preduce/3]).

merge_sort(List)
  when length(List) =< 1 ->
    List;
merge_sort(List) ->
  Mid = length(List) div 2,
    % split in two halves
  {L, R} = lists:split(Mid, List),
    % recursively sort each half
  SL = merge_sort(L),
  SR = merge_sort(R),
    % merge sorted halves
  merge(SL, SR).

merge([], R) -> R;
merge(L, []) -> L;
merge(L=[HL|TL], R=[HR|TR]) ->
  if
    HL =< HR -> [HL|merge(TL,R)];
    HL > HR  -> [HR|merge(L,TR)]
  end.

pmerge_sort(List)
  when length(List) =< 1 ->
    List;
pmerge_sort(List) ->
  Mid = length(List) div 2,
  {L, R} = lists:split(Mid, List),
  Pid = self(),
  spawn(fun ()-> Pid !
     {sl, pmerge_sort(L)} end),
  spawn(fun ()-> Pid !
     {sr, pmerge_sort(R)} end),
  receive {sl, SL} -> sl end,
  receive {sr, SR} -> sr end,
  merge(SL, SR).

% apply F to all
% elements of list
map(_, [])    -> [];
map(F, [H|T]) ->
  [F(H)|map(F,T)].

% parallel map
pmap(F, L) ->
  Me = self(), % my pid
  Ref = make_ref(),
    % for every E in L:
  Children = map(fun(E) ->
      % spawn a process
    spawn(fun() ->
      % sending Me result of F(E)
       Me ! {self(), Ref, F(E)}
          end) end, L),
    % collect and return results
  gather(Children, Ref).

% wait for all Children
% and collect results in order
gather(Children, Ref) ->
  [receive {Child, Ref, Res}
           -> Res end
     || Child <- Children].

reduce(_, A, []) -> A;
reduce(F, A, [H|T]) ->
  F(H, reduce(F, A, T)).

% parallel reduce
% for correctness:
%  - F is associative
%  - F(E, A) = F(A, E) = E
preduce(_, A, [])   -> A;
preduce(F, A, [E])  -> F(A, E);
preduce(F, A, List) ->
  Mid = length(List) div 2,
  {L, R} = lists:split(Mid, List),
  Me = self(), % L ++ R =:= Listn
  Lp = spawn(fun() -> % on left half
   Me ! {self(), preduce(F, A, L)} end),
  Rp = spawn(fun() -> % on right half
   Me ! {self(), preduce(F, A, R)} end),
  % combine results of left, right half
  F(receive {Lp, Lr} -> Lr end,
    receive {Rp, Rr} -> Rr end).

	 
	 
  
